from argparse import ArgumentParser
from contextlib import contextmanager
from pipes import quote
import subprocess
import sys
from textwrap import dedent

from yaml import safe_load

__metaclass__ = type


def parse_args(argv=None):
    """Parse the workspace-run arguments."""
    parser = ArgumentParser()
    parser.add_argument('config', help='Config file to use.')
    parser.add_argument('host', help='Machine to run the command on.')
    return parser.parse_args(argv)


class Primitives:
    """Base class for SSHPrimitives and FakePrimitives."""

    def __init__(self, host, workspace):
        self.host = host
        self.workspace = workspace

    @classmethod
    def create(cls, host):
        """Create an object of this class by making a tempdir on the host."""
        tempdir = cls.mktemp(host)
        return cls(host, tempdir)


class SSHPrimitives(Primitives):

    @classmethod
    def run_python(cls, host, script, argv=None):
        if argv is None:
            argv = []
        python_command = ' '.join(['python', '-'] + [quote(x) for x in argv])
        command = ['ssh', host, python_command]
        proc = subprocess.Popen(command, stdin=subprocess.PIPE,
                                stdout=subprocess.PIPE)
        output = proc.communicate(script)
        return output[0].rstrip('\n')

    @classmethod
    def mktemp(cls, host):
        """Make a temporary directory on the specified host via ssh."""
        return cls.run_python(host, dedent("""
            import tempfile
            print tempfile.mkdtemp(suffix=' foo')
            """))

    def destroy(self):
        """Destroy the workspace."""
        return self.run_python(self.host, dedent("""
            import shutil
            import sys
            shutil.rmtree(sys.argv[1])
            """), [self.workspace])

    def run(self, args, out_file):
        """Run a command in the workspace."""
        subprocess.check_call(
            ['ssh', self.host, 'cd', quote(self.workspace), ';'] + args,
            stdout=out_file)


@contextmanager
def workspace_context(host, primitives_factory):
    """Provide a context where a workspace is available.

    The primitives_factory is used to create a Primitives, and the resulting
    primitive is used to destroy it at the end.
    """
    primitives = primitives_factory.create(host)
    try:
        yield primitives
    finally:
        primitives.destroy()


def workspace_run(argv=None, primitives_factory=SSHPrimitives):
    """Run an operation in a workspace."""
    args = parse_args(argv)
    with open(args.config) as config_file:
        config = safe_load(config_file)
    with workspace_context(args.host, primitives_factory) as runner:
        runner.run(config['command'], sys.stdout)
