from argparse import Namespace
from contextlib import contextmanager
import os
from mock import patch
from pipes import quote
from shutil import rmtree
from StringIO import StringIO
from tempfile import (
    mkdtemp,
    NamedTemporaryFile,
    )
from textwrap import dedent
from unittest import TestCase
import subprocess

from yaml import safe_dump

from workspace_runner import (
    parse_args,
    Primitives,
    workspace_context,
    workspace_run,
    SSHPrimitives,
    )


class TestParseArgs(TestCase):

    def test_minimal(self):
        args = parse_args(['foo', 'bar'])
        self.assertEqual(args, Namespace(config='foo', host='bar'))


class FakePrimitives(Primitives):

    def __init__(self, host, workspace):
        super(FakePrimitives, self).__init__(host, workspace)
        self.run_calls = []

    @classmethod
    def mktemp(cls, host):
        return mkdtemp()

    def destroy(self):
        rmtree(self.workspace)

    def run(self, args, out_file):
        self.run_calls.append(args)
        out_file.write(''.join(args))


class TestPrimitives:

    def test_run(self):
        out_file = StringIO()
        with self.primitives_context() as primitives:
            self.expect_run(primitives, ['foo', 'bar'], out_file)
        self.assertEqual(out_file.getvalue(), 'foobar')

    def test_destroy(self):
        with self.create_context():
            primitives = self.primitives_class.create('asdf')
            with self.destroy_context(primitives):
                primitives.destroy()


class FakePrimitivesFactory:

    def __init__(self):
        self.last_instance = None

    def create(self, host):
        self.last_instance = FakePrimitives.create(host)
        return self.last_instance


class TestSSHPrimitives(TestPrimitives, TestCase):

    primitives_class = SSHPrimitives

    def setUp(self):
        super(TestSSHPrimitives, self).setUp()
        po_patcher = patch('subprocess.Popen', side_effect=Exception,
                           autospec=True)
        po_patcher.start()
        self.addCleanup(po_patcher.stop)

    @contextmanager
    def create_context(self):
        with patch('subprocess.Popen', autospec=True) as sp_mock:
            sp_mock.return_value.communicate.return_value = ('/tmp/workspace',)
            yield sp_mock
        sp_mock.assert_called_once_with(
            ['ssh', 'asdf', 'python -'], stdin=subprocess.PIPE,
            stdout=subprocess.PIPE)
        sp_mock.return_value.communicate.assert_called_once_with(dedent("""
            import tempfile
            print tempfile.mkdtemp(suffix=' foo')
            """))

    @contextmanager
    def destroy_context(self, primitives):
        with patch('subprocess.Popen', autospec=True) as po_mock:
            yield po_mock
        po_mock.assert_called_once_with(
            ['ssh', 'asdf', 'python - {}'.format(quote(primitives.workspace))],
            stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        po_mock.return_value.communicate.assert_called_once_with(dedent("""
            import shutil
            import sys
            shutil.rmtree(sys.argv[1])
            """))

    @contextmanager
    def primitives_context(self):
        with self.create_context():
            primitives = self.primitives_class.create('asdf')
            try:
                yield primitives
            finally:
                with self.destroy_context(primitives):
                    primitives.destroy()

    def expect_run(self, primitives, args, out_file):
        def effect(args, stdout):
            stdout.write(''.join(args[5:]))

        with patch('subprocess.check_call', side_effect=effect,
                   autospec=True) as cc_mock:
            primitives.run(args, out_file)
        call_args = ['ssh', primitives.host, 'cd',
                     quote(primitives.workspace), ';'] + args,
        cc_mock.assert_called_once_with(*call_args, stdout=out_file)

    def test_run_escape_workspace(self):
        out_file = StringIO()
        primitives = SSHPrimitives('host', 'my workspace')
        self.expect_run(primitives, ['foo', 'bar'], out_file)


class TestFakePrimitives(TestPrimitives, TestCase):

    primitives_class = FakePrimitives

    @contextmanager
    def create_context(self):
        yield

    @contextmanager
    def destroy_context(self, primitives):
        yield
        self.assertFalse(os.path.exists(primitives.workspace))

    @contextmanager
    def primitives_context(self):
        with workspace_context('asdf', self.primitives_class) as primitives:
            self.assertTrue(os.path.isdir(primitives.workspace))
            yield primitives
        self.assertFalse(os.path.exists(primitives.workspace))

    def expect_run(self, primitives, args, out_file):
        primitives.run(args, out_file)
        self.assertEqual(primitives.run_calls, [['foo', 'bar']])


class TestWorkspaceContext(TestCase):

    def test_minimal(self):
        with workspace_context('asdf', FakePrimitives) as primitives:
            self.assertTrue(os.path.exists(primitives.workspace))
        self.assertFalse(os.path.exists(primitives.workspace))


class TestWorkspaceRun(TestCase):

    def test_minimal(self):
        fp_factory = FakePrimitivesFactory()
        with NamedTemporaryFile() as config_file:
            safe_dump({'command': ['run', 'this']}, config_file)
            workspace_run([config_file.name, 'bar'],
                          fp_factory)
        primitives = fp_factory.last_instance
        self.assertEqual(primitives.run_calls, [['run', 'this']])
        self.assertFalse(os.path.exists(primitives.workspace))
